#!/usr/bin/perl
#
#    FEARLUS Expt Type 1 v. 0-5-1-5: create.pl
#    Copyright (C) 2007  Macaulay Institute
#
#    This file is part of FEARLUS Expt Type 1 v. 0-5-1-5, experiment scripts
#    for multiple runs with FEARLUS.
#
#    FEARLUS Expt Type 1 v. 0-5-1-5 is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    FEARLUS Expt Type 1 v. 0-5-1-5 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (LICENCE file in
#    this directory.)
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact information:
#      Gary Polhill,
#      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
#      g.polhill@macaulay.ac.uk
#
#
# create.pl
#
# Gary Polhill 14 June 2001
#
# This is a script to create experiments for comparing two subpopulations 
# in a number of different environments.

use strict;
use Cwd;
require "tools.pl";

# If requested, set up the comparison directories automatically. The -all
# option does this, and if given with the -self option, will include tests of
# sub-populations against themselves

if($#ARGV >= 0) {
  my $all = 0;
  my $self = 0;
  for(my $i = 0; $i <= $#ARGV; $i++) {
    if($ARGV[$i] eq "-all") {
      my @tmp = glob("*-v-*");     # sigh....
      if(scalar(@tmp) > 0) {
	warn "-all option ignored because files matching *-v-* exist in this".
	  " directory\n";
      }
      else {
	$all = 1;
      }
    }
    elsif($ARGV[$i] eq "-self") {
      $self = 1;
    }
    else {
      &usage("Invalid commandline argument: $ARGV[$i]");
    }
  }
  if($all) {
    my @subpops = glob("*.sp");
    for(my $i = 0; $i <= $#subpops; $i++) {
      substr($subpops[$i], -3) = "";
      if(!-e "$subpops[$i].ss") {
	die "Subpop file $subpops[$i].sp has no corresponding strategy ".
	  "selector file $subpops[$i].ss\n";
      }
    }
    my $selfstart = $self ? 0 : 1;
    for(my $i = 0; $i < $#subpops; $i++) {
      for(my $j = $i + $selfstart; $j <= $#subpops; $j++) {
	mkdir "$subpops[$i]-v-$subpops[$j]", 0777;
      }
    }
  }
  elsif($self) {
    warn "-self option ignored because no -all option or -all option ".
      "ignored\n";
  }
}

# Get the list of environments we are comparing the two subpopulations in

my @envs = glob("*.model");
my @shortenvs = &get_short_envs(@envs);

# Create a log file to save messages in

open(LOG, ">create.log") || die "Cannot create log file create.log: $!\n";

# Check that the climate and economy toggle files are present, check they are
# consistent with the model file, and read in the model parameters -- we'll
# need them to check what to hold constant in the paired replicate tests.

my %model;
my %warnings;

&load_environments(\@envs, \%model);

# Now loop through the subpopulations, setting up the experiments. There'll
# be one for each environment in each subpopulation contest directory.

foreach my $dir (glob("*-v-*")) {
  if(!-d "$dir") {
    print LOG "WARNING: $dir matches *-v-*, but is not a directory -- it will".
      " not be included in the experiment\n";
    next;
  }
  opendir(DIR, "$dir") || die "Could not open directory $dir";
  my @dirents = grep !/^\.\.?$/, readdir(DIR);
  closedir(DIR);
  if(scalar(@dirents) > 0) {
    print LOG "WARNING: $dir already has contents -- skipping to avoid any ".
      "over-writes\n";
    next;
  }
  $dir =~ /^(.*)-v-(.*)$/;
  my $sp1 = $1;
  my $sp2 = $2;
  &check_subpops(\@envs, \%model, $sp1, $sp2);
  # Loop through all the environments
  for(my $env = 0; $env <= $#envs; $env++) {
    # Create the environment directory
    my $envdir = "$shortenvs[$env]";
    mkdir "$dir/$envdir", 0777 ||
      die "Could not create directory $dir/$envdir: $!\n";
    print LOG "NOTE: Created directory $dir/$envdir\n";
    # Create the top-level subpopulation file
    my $spf = &create_subpop_file($dir, $envdir, [$sp1, $sp2]);
    # Create the report configuration file
    open(REPCFG, ">$dir/$envdir/Parcel.repcfg") ||
      die "Could not create configuration file $dir/$envdir/Parcel.repcfg: ".
	"$!\n";
    my $reportYear = $model{$envs[$env], 'maxYear'} - 1;
    print REPCFG "ParcelSubPopReport YearsToReport: $reportYear\n";
    print REPCFG "End\n";
    close(REPCFG);
    # Copy the subpopulation and environment support files
    my @spfiles = ($sp1 eq $sp2) ? ("${sp1}.sp", "${sp1}.ss") 
      : ("${sp1}.sp", "${sp2}.sp", "${sp1}.ss", "${sp2}.ss");
    foreach my $file (@spfiles,
		      $model{$envs[$env], 'climateToggleProbFile'},
		      $model{$envs[$env], 'economyToggleProbFile'}) {
      &cp($file, "$dir/$envdir/$file");
    }
    # Create the model file. Some model parameters are written automatically,
    # others are as per that in the top level directory -- can't notify the 
    # user of this or they'll be flooded with messages...
    my $modelfile = $envs[$env];
    open(MODEL, ">$dir/$envdir/$modelfile") || 
      die "Cannot create model file $dir/$envdir/$modelfile: $!\n";
    print MODEL '@begin', "\n";
    # First write all the automatic parameters
    foreach my $autoparam (['nSubPops', 2],
			   ['subPopFile', "$spf"],
			   ['infiniteTime', 0]) {
      my($param, $value) = @$autoparam;
      print MODEL "$param\t$value\n";
      if(defined($model{$modelfile, $param})
	 && $model{$modelfile, $param} ne $value) {
	print LOG "WARNING: Over-riding value of $param in modelfile ".
	  "$modelfile ($model{$modelfile, $param}) with value $value\n"
	    unless defined($warnings{$modelfile, $param, 'override'});
	$warnings{$modelfile, $param, 'override'} = $value;
      }
    }
    # Write the paired replicate parameters. There are none.
    print MODEL "useLandParcelFile\t0\n";
    print MODEL "useClimateFile\t0\n";
    print MODEL "useEconomyFile\t0\n";
    print MODEL "useLandUseFile\t0\n";
    # Write out the rest of the parameters. Where these have not been
    # specified, either default values will be used, or an error message
    # will be generated.
    # Parameters with default values first
    foreach my $paramdefault (['neighbourhoodRadius', 1],
			      ['nInitXParcels', 1],
			      ['nInitYParcels', 1],
			      ['strategyChangeUnit', 0.0],
			      ['neighbourNoiseMax', 0.0],
			      ['neighbourNoiseMin', 0.0]) {
      my($param, $default) = @$paramdefault;
      if(defined($model{$modelfile, $param})) {
	print MODEL "$param\t$model{$modelfile, $param}\n";
      }
      else {
	print MODEL "$param\t$default\n";
	print LOG "WARNING: $modelfile has no value for $param, using ".
	  "default value ($default)\n" 
	    unless defined($warnings{$modelfile, $param, 'default'});
	$warnings{$modelfile, $param, 'default'} = $default;
      }
    }
    # Now the rest of the parameters
    foreach my $param ('environmentType',
		       'nLandUse',
		       'pLandUseDontCare',
		       'clumping',
		       'envXSize',
		       'envYSize',
		       'breakEvenThreshold',
		       'landParcelPrice',
		       'climateBSSize',
		       'economyBSSize',
		       'landParcelBSSize',
		       'climateToggleProbFile',
		       'economyToggleProbFile',
		       'maxYear'
		      ) {
      if(defined($model{$modelfile, $param})) {
	print MODEL "$param\t$model{$modelfile, $param}\n";
      }
      else {
	die "Cannot find parameter $param in model file $modelfile\n";
      }
    }
    # Finish off the file
    print MODEL '@end', "\n";
    close(MODEL);
  }
}
close(LOG);

exit 0;

sub usage {
  my ($message) = @_;

  die <<DEATH;
Error: $message

Usage: $0 [-all [-self]]

Create the files needed to run a series of experiments comparing one
subpopulation against another in a number of environments. If you give the -all
option it will create contests of all subpopulations in the current working
directory against all the other subpopulations in the cwd, provided there are
no files in the directory that match *-v-*. If you additionally give the -self
option, then directories will be created pitting subpopulations against 
themselves.
DEATH
}
