#!/usr/bin/perl
#
#    FEARLUS Expt Type 1 v. 0-5-1-5: results.pl
#    Copyright (C) 2007  Macaulay Institute
#
#    This file is part of FEARLUS Expt Type 1 v. 0-5-1-5, experiment scripts
#    for multiple runs with FEARLUS.
#
#    FEARLUS Expt Type 1 v. 0-5-1-5 is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    FEARLUS Expt Type 1 v. 0-5-1-5 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (LICENCE file in
#    this directory.)
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact information:
#      Gary Polhill,
#      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
#      g.polhill@macaulay.ac.uk
#
#
# results.pl
#
# Gary Polhill 15 June 2001
#
# This is a script to analyse the results of experiments comparing two
# subpopulations in a number of different environments. A subpopulation is
# viewed as superior to another in a particular environment if it ends up with
# the majority of land parcels in a significant number of runs using the
# binomial test.
#
# The script takes as argument a list of significance levels. It may optionally
# be given the argument -histo, which will count up all the runs in a
# particular directory and add together all the land parcels that each subpop
# had in the final year of each run, and draw a histogram of the result.

use strict;
use Cwd;
require "tools.pl";

my $histo = 0;
my (@histo1, @histo2);
my $tails = 1;

# Check the arguments to the program

while($ARGV[0] =~ /^-/) {
  my $option = shift(@ARGV);
  if($option eq '-histo') {
    $histo = shift(@ARGV);
  }
  elsif($option eq '-2tail') {
    $tails = 2;
  }
  else {
    &usage("Invalid option: $option");
  }
}

my @siglevels = @ARGV;
if(scalar(@siglevels) == 0) {
  &usage("You must supply at least one significance level");
}

# Get the list of environments in which the subpopulations are being compared

my @envs = glob("*.model");
my @shortenvs = &get_short_envs(@envs);

# Create a log file to save the results in.

open(LOG, ">results.log") || die "Cannot create log file results.log: $!\n";

# Loop through the subpopulation directories and check the results

foreach my $spdir (glob("*-v-*")) {
  # Ignore non-directories that match the above glob
  next if !-d "$spdir";
  # Pull out the subpop labels from the directory name
  $spdir =~ /^(.*)-v-(.*)$/;
  my $sp1 = $1;
  my $sp2 = $2;
  my $sp1summaryflag = 0;
  my $sp2summaryflag = 0;
  my $sp1summary;
  my $sp2summary;
  # Loop through all the environments
  for(my $env = 0; $env <= $#shortenvs; $env++) {
    my $envdir = "$shortenvs[$env]";
    # Check the environment contest sub-directory exists
    if(!-d "$spdir/$envdir") {
      die "Cannot find directory $spdir/$envdir -- have you run the ".
	"creation script (cwd = ".getcwd().")?\n";
    }
    # Change into the experiment contest sub-directory
    chdir "$spdir/$envdir" 
      || die "Cannot change directory to $spdir/$envdir: $!\n";
    # Declare a hash to use to store the seeds (all the runs in a particular
    # directory must have different seeds)
    my %seed;
    # Define variables to store the total number of times each subpopulation
    # did better than the other in this environment
    my($sp1win, $sp2win) = (0, 0);
    my($sp1eq, $sp2eq) = (0, 0);
    # Define a variable to store the total number of valid runs -- this is
    # where the subpopulations match and the seeds are all different
    my $validruns = 0;
    # Initialise the histogram variables
    my $parceltot1 = 0;
    my $parceltot2 = 0;
    for(my $i = 0; $i <= $#histo1; $i++) {
      $histo1[$i] = 0;
    }
    for(my $i = 0; $i <= $#histo2; $i++) {
      $histo2[$i] = 0;
    }
    # Loop through all the runs
    my $year;
    my $run;
  RUN:
    for($run = 1;
	-f (my $reportfile = "$spdir-$shortenvs[$env]-run$run-rep.txt");
	$run++) {
      # Load in the report file
      my %report;
      &load_report_file($reportfile, \%report);
      # Check seeds
      my $flag = 0;
      if(defined($seed{$report{'Parameter', 'Seed for simulation'}})) {
	print LOG "WARNING: Seed ".
	  "$report{'Parameter', 'Seed for simulation'} used in ".
	    "$seed{$report{'Parameter', 'Seed for simulation'}} ".
	      "and $spdir/$envdir/$reportfile\n";
	$flag = 1;
	next;
      }
      $seed{$report{'Parameter', 'Seed for simulation'}} =
	"$spdir/$envdir/reportfile";
      next if $flag;
      # Check sub-populations -- all runs must use the same subpopulations
      # The create script makes sure that the subpopulation file is
      # consistent, so the subpopulations will always be in the same order.
      foreach my $subpop ([$sp1, \$sp1summaryflag, \$sp1summary, 1],
			  [$sp2, \$sp2summaryflag, \$sp2summary, 2]) {
	my($sp, $summaryflag, $summary, $id) = @$subpop;
	if(!$$summaryflag) {
	  $$summaryflag = 1;
	  $$summary = $report{'SubpopSummary', $id};
	  print LOG "NOTE: $sp is $$summary in $spdir/*\n";
	}
	elsif($$summary ne $report{'SubpopSummary', $id}) {
	  print LOG "WARNING: Inconsistencies for $sp -- not the same in ".
	    "previous runs as in $spdir/$envdir/$reportfile\n\tPreviously:".
	      "\t$$summary\n\tCurrently:\t$report{'SubpopSummary', $id}\n";
	  next RUN;
	}
      }
      # Find out the year of the report and check they are the same in each run
      if($run == 1 || !defined($year)) {
	$year = pop(@{$report{'ParcelSubPopReport', 'ReportingYears'}});
	if(!defined($year)) {
	  print LOG "WARNING: Run $run has no ParcelSubPopReports in ".
	    "$spdir/$envdir\n" if $run == 1;
	  next RUN;
	}
      }
      elsif($year != pop(@{$report{'ParcelSubPopReport', 'ReportingYears'}})) {
	print LOG "WARNING: Run $run reports on a different year for ".
	  "ParcelSubPopReport than that in run 1 in".
	    " $spdir/$envdir, run $run\n";
	next;
      }
      # OK, so now, who did better than whom?
      # Have to use temporary variables to store the associative arrays
      # dereferenced from the report for programmer sanity -- XEmacs can't
      # cope with the syntax for indenting purposes (even though there's
      # no ${!!)
      my %rsp = %{$report{'ParcelSubPopReport', $year}};
      # So now use $rsp{1} instead of
      # ${$report{'ParcelSubPopReport', $year}}{1}
      $sp1win++ if($rsp{1} > $rsp{2});
      $sp2win++ if($rsp{2} > $rsp{1});
      $sp1eq++ if($rsp{1} == $rsp{2});
      $sp2eq++ if($rsp{1} == $rsp{2});
      # Increment the number of valid runs
      $validruns++;
      # Add the results to the histogram
      if($histo > 0) {
	$histo1[int($rsp{1} / $histo)]++;
	$histo2[int($rsp{2} / $histo)]++;
	$parceltot1 += $rsp{1};
	$parceltot2 += $rsp{2};
      }
    }
    # Print the p-value for the results in this run using a 1-tailed
    # binomial test 
    if($validruns > 0) {
      if($tails == 1) {
	print "Results for $spdir/$envdir (1-tailed binomial test of ".
	  "$validruns valid runs)\n";
	print "Hypothesis\tSignificance\tp-value\tDraws\tWins\t".
	  "Non-tied runs\n";
	foreach my $test ([$sp1, $sp2, $env, $sp1win, $sp1eq],
			  [$sp2, $sp1, $env, $sp2win, $sp2eq]) {
	  my($sp, $vsp, $inenvref, $score, $nodiff) = @$test;
	  my $inenv = $shortenvs[$inenvref];
	  my $pvalue;
	  # Find the smallest of the command-line entered p-values that
	  # is greater than this p-value.
	  my $significance = "NS";
	  my $minlevel = 1;
	  foreach my $userlevel (@siglevels) {
	    my $significant = &testgtbinomial($validruns - $nodiff, $score,
					      $userlevel, \$pvalue);
	    if($significant && $userlevel < $minlevel) {
	      $minlevel = $userlevel;
	      $significance = $userlevel;
	    }
	  }
	  print "$sp does better against $vsp in environment $inenv".
	    "\t$significance\t$pvalue\t$nodiff\t$score\t",
	    $validruns - $nodiff, "\n";
	}
      }
      elsif($tails == 2) {
	print "Results for $spdir/$envdir (2-tailed binomial test of ".
	  "$validruns valid runs)\n";
	print "Hypothesis\tSignificance\tp-value\t$sp1 wins\tDraws\t$sp2 wins".
	  "\tNon-tied runs\n";
	my $inenv = $shortenvs[$env];
	my $pvalue;
	# Find the smallest of the command-line entered p-values that is
	# greater than this p-value.
	my $significance = "NS";
	my $minlevel = 1;
	foreach my $userlevel (@siglevels) {
	  my $significant = &testnebinomial($validruns - $sp1eq, $sp1win,
					    $userlevel, \$pvalue);
	  if($significant && $userlevel < $minlevel) {
	    $minlevel = $userlevel;
	    $significance = $userlevel;
	  }
	}
	if($sp1win > $sp2win) {
	  print "$sp1 and $sp2 are not equal in environment $inenv";
	}
	else {
	  print "$sp2 and $sp1 are not equal in environment $inenv";
	}
	print "\t$significance\t$pvalue\t$sp1win\t$sp1eq\t$sp2win\t",
	$validruns - $sp1eq, "\n";
      }
      if($histo > 0) {
	foreach my $histogram ([$sp1, $parceltot1, \@histo1],
			       [$sp2, $parceltot2, \@histo2]) {
	  my($sp, $parceltot, $histarr) = @$histogram;
	  print "Total parcels for $sp over $validruns runs:\t$parceltot\t".
	    "average:\t", ($parceltot / $validruns), "\n";
	  print "Distribution for $sp parcels:\n";
	  for(my $i = 0; $i <= $#$histarr; $i++) {
	    print (($histo * $i), " - ", (($histo * ($i + 1)) - 1), "\t",
	    ('|' x $$histarr[$i]), "\t$$histarr[$i]\n");
	  }
	}
      }
    }
    else {
      print "There were no valid runs in $spdir/$envdir\n";
    }
    # Change directory back to cwd ready for next set of runs
    chdir "$ENV{PWD}";
  }
}
close(LOG);
print "Results completed. Check results.log for any messages\n";
exit 0;

sub usage {
  my($message) = @_;
  die <<DEATH;
$message

This script is called as follows:
$0 [-histo n] [-2tail] <list of significance levels>

This program looks in the report files of the runs completed and does a
binomial sign test at the significance levels specified to see if one
sub-population does better than another in a number of different environments.

It may optionally be given -histo n, in which case results will be displayed
that show the total number of land parcels that each subpopulation had at the
end of all the runs, along with a histogram (with bucket size n) showing the
distribution of the total.

You may also optionally specify the -2tail option, which will do a two-tailed
test instead of a one-tailed.
DEATH
}
